/* mpn_modexact_1c_odd - from x86/fat directory for fat binary.
   Generated by configure - DO NOT EDIT. */

#define OPERATION_mode1o 1
#define __gmpn_modexact_1c_odd           __gmpn_modexact_1c_odd_fat
#define __gmpn_modexact_1cc_odd          __gmpn_modexact_1cc_odd_fat
#define __gmpn_preinv_modexact_1c_odd  __gmpn_preinv_modexact_1c_odd_fat
#define __gmpn_modexact_1c_odd_cps     __gmpn_modexact_1c_odd_cps_fat

#include ".././mpn/x86/fat/mode1o.c"

