/* mpn_com - from x86/fat directory for fat binary.
   Generated by configure - DO NOT EDIT. */

#define OPERATION_com 1
#define __gmpn_com           __gmpn_com_fat
#define __gmpn_comc          __gmpn_comc_fat
#define __gmpn_preinv_com  __gmpn_preinv_com_fat
#define __gmpn_com_cps     __gmpn_com_cps_fat

#include ".././mpn/x86/fat/com.c"

