/* mpn_redc_1 - from x86/fat directory for fat binary.
   Generated by configure - DO NOT EDIT. */

#define OPERATION_redc_1 1
#define __gmpn_redc_1           __gmpn_redc_1_fat
#define __gmpn_redc_1c          __gmpn_redc_1c_fat
#define __gmpn_preinv_redc_1  __gmpn_preinv_redc_1_fat
#define __gmpn_redc_1_cps     __gmpn_redc_1_cps_fat

#include ".././mpn/x86/fat/redc_1.c"

