/* mpn_mullo_basecase - from x86/fat directory for fat binary.
   Generated by configure - DO NOT EDIT. */

#define OPERATION_mullo_basecase 1
#define __gmpn_mullo_basecase           __gmpn_mullo_basecase_fat
#define __gmpn_mullo_basecasec          __gmpn_mullo_basecasec_fat
#define __gmpn_preinv_mullo_basecase  __gmpn_preinv_mullo_basecase_fat
#define __gmpn_mullo_basecase_cps     __gmpn_mullo_basecase_cps_fat

#include ".././mpn/x86/fat/mullo_basecase.c"

