/* mpn_mod_1s_2p - from x86/fat directory for fat binary.
   Generated by configure - DO NOT EDIT. */

#define OPERATION_mod_1_2 1
#define __gmpn_mod_1s_2p           __gmpn_mod_1s_2p_fat
#define __gmpn_mod_1cs_2p          __gmpn_mod_1cs_2p_fat
#define __gmpn_preinv_mod_1s_2p  __gmpn_preinv_mod_1s_2p_fat
#define __gmpn_mod_1s_2p_cps     __gmpn_mod_1s_2p_cps_fat

#include ".././mpn/x86/fat/mod_1_2.c"

