/* mpn_lshiftc - from x86/fat directory for fat binary.
   Generated by configure - DO NOT EDIT. */

#define OPERATION_lshiftc 1
#define __gmpn_lshiftc           __gmpn_lshiftc_fat
#define __gmpn_lshiftcc          __gmpn_lshiftcc_fat
#define __gmpn_preinv_lshiftc  __gmpn_preinv_lshiftc_fat
#define __gmpn_lshiftc_cps     __gmpn_lshiftc_cps_fat

#include ".././mpn/x86/fat/lshiftc.c"

