/* mpn_redc_2 - from x86/fat directory for fat binary.
   Generated by configure - DO NOT EDIT. */

#define OPERATION_redc_2 1
#define __gmpn_redc_2           __gmpn_redc_2_fat
#define __gmpn_redc_2c          __gmpn_redc_2c_fat
#define __gmpn_preinv_redc_2  __gmpn_preinv_redc_2_fat
#define __gmpn_redc_2_cps     __gmpn_redc_2_cps_fat

#include ".././mpn/x86/fat/redc_2.c"

