/* mpn_mod_1_1p - from x86/fat directory for fat binary.
   Generated by configure - DO NOT EDIT. */

#define OPERATION_mod_1_1 1
#define __gmpn_mod_1_1p           __gmpn_mod_1_1p_fat
#define __gmpn_mod_1c_1p          __gmpn_mod_1c_1p_fat
#define __gmpn_preinv_mod_1_1p  __gmpn_preinv_mod_1_1p_fat
#define __gmpn_mod_1_1p_cps     __gmpn_mod_1_1p_cps_fat

#include ".././mpn/x86/fat/mod_1_1.c"

